use std::collections::{BTreeMap, HashMap};

use crate::ast::Node as AstNode;

#[derive(Debug)]
pub struct SymbolRecord {
    name: String,
    decl: AstNode,
}

impl SymbolRecord {
    pub fn node(&self) -> AstNode {
        self.decl
    }

    pub fn name(&self) -> &str {
        &self.name
    }
}

pub enum SymbolKind {
    Var, // or Let
    Orderless,
}

#[derive(Debug, Default)]
pub struct SymbolTable {
    // this is a `Vec<_>` because order matters. Some symbols such as functions
    // cannot be shadowed, but I really like shadowing variables and function
    // parameters, so any `x` may be redefined.
    ordered_identifiers: Vec<SymbolRecord>,
    orderless_identifiers: HashMap<String, SymbolRecord>,
    parent: Option<Box<SymbolTable>>,
    children: BTreeMap<Option<AstNode>, SymbolTable>,
    scope: Option<AstNode>,
}

impl SymbolTable {
    pub fn new() -> SymbolTable {
        Self {
            ..Default::default()
        }
    }

    pub fn root_ref(&self) -> &SymbolTable {
        match self.parent.as_ref() {
            Some(parent) => parent.root_ref(),
            None => self,
        }
    }

    pub fn root_mut(&mut self) -> &mut SymbolTable {
        let this = self as *mut Self;
        unsafe {
            match (&mut *this).parent.as_mut() {
                Some(parent) => parent.root_mut(),
                None => self,
            }
        }
    }

    pub fn parent_ref(&self) -> &SymbolTable {
        match self.parent.as_ref() {
            Some(parent) => Box::as_ref(parent),
            None => self,
        }
    }

    pub fn parent_mut(&mut self) -> &mut SymbolTable {
        let this = self as *mut Self;
        unsafe {
            match (&mut *this).parent.as_mut() {
                Some(parent) => Box::as_mut(parent),
                None => self,
            }
        }
    }

    pub fn insert_symbol(&mut self, name: &str, node: AstNode, kind: SymbolKind) -> &SymbolRecord {
        match kind {
            SymbolKind::Var => {
                self.ordered_identifiers.push(SymbolRecord {
                    name: name.to_owned(),
                    decl: node,
                });
                self.ordered_identifiers.last().unwrap()
            }
            _ => {
                self.orderless_identifiers.insert(
                    name.to_owned(),
                    SymbolRecord {
                        name: name.to_owned(),
                        decl: node,
                    },
                );
                self.orderless_identifiers.get(name).unwrap()
            }
        }
    }

    pub fn insert_orderless_symbol(&mut self, name: &str, node: AstNode) -> &SymbolRecord {
        self.orderless_identifiers.insert(
            name.to_owned(),
            SymbolRecord {
                name: name.to_owned(),
                decl: node,
            },
        );
        self.orderless_identifiers.get(name).unwrap()
    }

    pub fn find_symbol_or_insert_with<'a, F>(&'a mut self, name: &str, cb: F) -> &'a SymbolRecord
    where
        F: FnOnce() -> (AstNode, SymbolKind),
    {
        let this = self as *mut Self;
        if let Some(record) = unsafe { &*this }.find_symbol(name) {
            record
        } else {
            let (node, kind) = cb();
            self.insert_symbol(name, node, kind)
        }
    }

    pub fn find_symbol_by_decl(&self, decl: AstNode) -> Option<&SymbolRecord> {
        self.ordered_identifiers
            .iter()
            .find(|r| r.decl == decl)
            .or_else(|| {
                self.orderless_identifiers
                    .iter()
                    .find(|(_, v)| v.decl == decl)
                    .map(|(_, v)| v)
            })
            .or_else(|| {
                self.parent
                    .as_ref()
                    .and_then(|p| p.find_symbol_by_decl(decl))
            })
    }

    pub fn find_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.ordered_identifiers
            .iter()
            .find(|r| r.name.as_str() == name)
            .or_else(|| self.orderless_identifiers.get(name))
            .or_else(|| self.parent.as_ref().and_then(|p| p.find_symbol(name)))
    }

    pub fn find_orderless_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.orderless_identifiers.get(name).or_else(|| {
            self.parent
                .as_ref()
                .and_then(|p| p.find_orderless_symbol(name))
        })
    }

    pub fn into_find_child(&mut self, scope: AstNode) -> Option<()> {
        if let Some(mut parent) = self.children.remove(&Some(scope)) {
            core::mem::swap(self, &mut parent);
            self.parent = Some(Box::new(parent));
            Some(())
        } else {
            None
        }
    }

    pub fn into_child(&mut self, scope: AstNode) {
        let mut parent = Self {
            scope: Some(scope),
            ..Default::default()
        };
        core::mem::swap(self, &mut parent);
        self.parent = Some(Box::new(parent));
    }

    pub fn extend_orderless<I>(&mut self, iter: I)
    where
        I: IntoIterator<Item = (String, SymbolRecord)>,
    {
        self.orderless_identifiers.extend(iter)
    }

    pub fn extract_orderless_if<F>(
        &mut self,
        pred: F,
    ) -> std::collections::hash_map::ExtractIf<String, SymbolRecord, F>
    where
        F: FnMut(&String, &mut SymbolRecord) -> bool,
    {
        self.orderless_identifiers.extract_if(pred)
    }

    /// returns `self` if `self.parent` was `Some(_)`.
    pub fn into_parent(&mut self) {
        if let Some(child) = self.parent.take() {
            let mut child = Box::into_inner(child);
            core::mem::swap(self, &mut child);
            self.children.insert(child.scope, child);
        }
    }
}
